"""Arguments for the graphical user interface."""


def add_argparser_arguments(parser):
    parser.add_argument(
        "-W", "--window-width",
        help="set window width",
        action="store",
        type=int,
        default=500,
    )
    parser.add_argument(
        "-H", "--window-height",
        help="set window height",
        action="store",
        type=int,
        default=400,
    )



import glob
import os

import yaml

from meety import io
from meety.logging import log
from meety.meetings.meeting import Meeting


class Loader:
    """Manages file selection and loading.

    There are implicit directories, provided by default, and explicit
    directories, provided by the user. The loader will consider all YAML files
    (by extension) in the explicit directories and, if not disabled, also in
    the implicit directories.
    Additionally, the loader will consider all explicitly specified YAML files.
    """

    implicit_directories = [
        os.getcwd(),
        os.path.expanduser("~"),
    ]

    @classmethod
    def _yaml_files_in_directories(cls, directories):
        """Returns paths to all files with extension 'yaml' or 'yml' in
        the given list of directories.
        """
        files = []
        for directory in directories:
            files.extend(cls._yaml_files_in_directory(directory))
        return files

    @classmethod
    def _yaml_files_in_directory(cls, directory):
        """Returns paths to all files with extension 'yaml' or 'yml' in
        the given directory.
        """
        def files_in_by_extension(directory, ext):
            return glob.glob(os.path.join(directory, f"*.{ext}"))
        return (
            files_in_by_extension(directory, "yaml")
            + files_in_by_extension(directory, "yml")
        )

    @classmethod
    def _load_meeting_entries_from_file(cls, filename):
        """Try to load meetings from YAML file, return """
        data = cls._load_data_from_file(filename)
        return cls._create_meetings_from_data(data)

    @classmethod
    def _load_data_from_file(cls, filename):
        """Try to load data from YAML file, return empty dictionary on
        failure.
        """
        try:
            with open(filename) as file:
                return yaml.load(file, Loader=yaml.BaseLoader)
        except FileNotFoundError as e:
            log.warning(f"Meeting file '{filename}' does not found.")
            log.exception(e)
            return {}
        except Exception as e:
            log.warning(f"Failed to load meetings from '{filename}'")
            log.exception(e)
            return {}

    @classmethod
    def _create_meetings_from_data(cls, data):
        """Create meeting from a single entry or meetings from a list
        of entries."""
        meetings = []
        if not isinstance(data, list):
            data = [data]
        for d in data:
            m = cls._create_meeting_from_data(d)
            if m:
                meetings.append(m)
        return meetings

    @classmethod
    def _create_meeting_from_data(cls, data):
        """Create a meeting from a single entry."""
        clean_data = io.process(data)
        if Meeting.is_meeting_entry(clean_data):
            return Meeting(data, clean_data)

    def __init__(self):
        """Start with no explicit directories and files but with
        implicit directories enabled."""
        self._explicit_directories = []
        self._explicit_files = []
        self.only_explicit = False
        self._loaded_paths = LoadedPaths()
        self._meetings = []

    def add_explicit_directories(self, directories):
        self._explicit_directories.extend(directories)

    def add_explicit_files(self, files):
        self._explicit_files.extend(files)

    @property
    def active_directories(self):
        if self.only_explicit:
            return self._explicit_directories
        return self._explicit_directories + Loader.implicit_directories

    @property
    def explicit_files(self):
        return self._explicit_files

    @property
    def implicit_files(self):
        return Loader._yaml_files_in_directories(self.active_directories)

    @property
    def all_files(self):
        return self.explicit_files + self.implicit_files

    @property
    def loaded_paths(self):
        return self._loaded_paths

    @property
    def meetings(self):
        return self._meetings

    def load(self):
        """Load from all explicitly and implicitly specified files if
        the file has not already been read.
        """
        files = self.all_files
        self._log_files_to_consider(files)
        for f in files:
            self._consider_to_load_from_file(f)

    def reload(self):
        """Reload data from already loaded files."""
        self._meetings = []
        for path in self._loaded_paths.all_paths():
            self.load_from_file(path)

    def _log_files_to_consider(self, files):
        if not files:
            log.warning("No YAML files to consider!")
        else:
            log.info("Will consider the following YAML files:")
            log.info("\n".join([f"  - {f}" for f in files]))

    def _consider_to_load_from_file(self, filename):
        """Load from file if it has not already been loaded."""
        if self._loaded_paths.contains(filename):
            log.info(f"Skipping file '{filename}', already read.")
        else:
            self.load_from_file(filename)
            self._loaded_paths.add(filename)

    def load_from_file(self, filename):
        """Unconditionally load meetings from file and save them."""
        meetings = self._load_meeting_entries_from_file(filename)
        new_meetings = [m for m in meetings if m not in self._meetings]
        self._log_loading_stats(filename, meetings, new_meetings)
        self._meetings.extend(new_meetings)

    def _log_loading_stats(self, filename, meetings, new_meetings):
        num_read = len(meetings)
        num_new = len(new_meetings)
        details = [
            f"Adding only {num_new} new.",
            "Adding all.",
        ][num_read == num_new]
        log.info(
            f"Loaded {num_read} meetings from '{filename}'. "
            + details
        )


class LoadedPaths:
    """Remember absolute paths of files that have already been loaded."""
    def __init__(self):
        self._abspaths = set()

    def contains(self, path):
        abspath = os.path.abspath(path)
        if abspath in self._abspaths:
            return True
        else:
            return False

    def add(self, path):
        abspath = os.path.abspath(path)
        self._abspaths.add(abspath)

    def all_paths(self):
        return list(self._abspaths)



from fnmatch import fnmatch

from meety.io.actions import registry
from meety.io.actions.action import Action


class SynonymAction(Action):
    """Simplified version of `infer_action`, providing an entry under a
    new key.

    For example, `infer_action({"user": "JohnDoe"}, "name", ["user"])`
    replaces the key "user" by "name".
    """

    name = "synonym"

    def __init__(self, output, inputs):
        super().__init__()
        self._output = output
        self._inputs = inputs

    def is_applicable(self, data, current_level):
        return (
            super().is_applicable(data, current_level)
            and self._output not in data.keys()
        )

    def _apply(self, data):
        try:
            key = next(self._matching_keys(data))
        except StopIteration:
            pass
        else:
            data[self._output] = data[key]
            data.pop(key)
        finally:
            return data

    def _matching_keys(self, data):
        inputs = [inp.casefold() for inp in self._inputs]
        for input_pattern in inputs:
            for k in data.keys():
                if fnmatch(k, input_pattern):
                    yield k

    @classmethod
    def create(cls, data):
        output = data["output"]
        inputs = data["inputs"]
        return SynonymAction(output, inputs)

    @classmethod
    def register(cls):
        registry.register_action_class("synonym", SynonymAction)



"""Arguments for the command-line interface."""


def add_argparser_arguments(parser):
    parser.add_argument(
        "-a", "--all",
        help="include all meetings",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-t", "--test-run",
        help="print connect command, don't connect",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-1", "--first",
        help="choose first if there are multiple meetings",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-b", "--best",
        help="consider only best matching meetings",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-y", "--yes",
        help="assume yes (don't ask to connect)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-s", "--show-only",
        help="show only matches, don't connect",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-c", "--choose-connection",
        help="ask for connection handler, if multiple",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--list-connection-handlers",
        help="list only active connection handlers",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--list-meetings",
        help="list only meeting details",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--list-ratings",
        help="list only ratings",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--datetime",
        help="use given datetime instead of current",
        action="store",
        default=None,
    )
    parser.add_argument(
        "query",
        help="search phrase",
        nargs="*",
        default=["*"],
    )



from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QPushButton,
    QWidget,
)

from meety.gui.info_dialog import InfoDialog


class StatusWidget(QWidget):
    NOTIFY_QUICK_TIME_SPAN = 3000
    NOTIFY_NORMAL_TIME_SPAN = 8000

    def __init__(self):
        super().__init__()
        self._message_queue = []
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()
        self.show_default_message()

    def _init_style(self):
        pass

    def _add_widgets(self):
        layout = self._create_layout()
        layout.addWidget(self._create_status(), 1)
        layout.addWidget(self._create_info(), 0)
        self.setLayout(layout)

    def _create_layout(self):
        layout = QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        return layout

    def _create_status(self):
        self._status = QLabel("")
        return self._status

    def _create_info(self):
        self._info = QPushButton("&info")
        self._info.setToolTip(
            "Show logs, loaded data and version information."
        )
        self._info.setProperty("class", "info")
        return self._info

    def _connect_widget_signals(self):
        self._info.clicked.connect(self.show_info)
        self._timer = QTimer()
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self._show_next_message)

    def show_default_message(self):
        self._status.setText("Ready.")

    def notify_quickly(self, text):
        span = self.NOTIFY_QUICK_TIME_SPAN
        self._enqueue_notification(text, "", span)

    def notify(self, text, tooltip=""):
        span = self.NOTIFY_NORMAL_TIME_SPAN
        self._enqueue_notification(text, tooltip, span)

    def _enqueue_notification(self, text, tooltip, time_span):
        self._message_queue.append((text, tooltip, time_span))
        if len(self._message_queue) == 1:
            self._show_message()

    def _show_next_message(self):
        self._message_queue.pop(0)
        if self._message_queue:
            self._show_message()
        else:
            self.show_default_message()

    def _show_message(self):
        text, tooltip, time_span = self._message_queue[0]
        self._status.setText(text)
        self._status.setToolTip(tooltip)
        self._timer.start(time_span)

    def clear_notifications(self):
        if self._message_queue:
            self._timer.stop()
            self._message_queue.clear()
            self.show_default_message()

    def show_info(self):
        dialog = InfoDialog()
        dialog.exec_()



from abc import (
    ABC,
    abstractmethod,
)
from fnmatch import fnmatch

from meety.logging import log


class Action(ABC):
    def __init__(self):
        self._level = "*"

    def set_conditions(self, level_pattern):
        self._level = level_pattern

    def set_conditions_from_dict(self, data):
        self.set_conditions(data.get("level", "*"))

    def satisfies_conditions(self, current_level):
        return (
            current_level is None
            or fnmatch(str(current_level), self._level)
        )

    def is_applicable(self, data, current_level):
        return (
            isinstance(data, dict)
            and self.satisfies_conditions(current_level)
        )

    def apply(self, data, current_level=None):
        if not self.is_applicable(data, current_level):
            return data
        else:
            return self._do_apply(data)

    def _do_apply(self, data):
        try:
            self._apply(data)
        except KeyError as e:
            log.warning(
                f"Failed to apply action '{data}',"
                f" missing attribute '{e}'."
            )
        except Exception as e:
            log.warning(f"Failed to apply action '{data}'.")
            log.exception(e)
        finally:
            return data

    @abstractmethod
    def _apply(self, data):
        pass



"""MacOS implementation for system-dependent functionality."""

import subprocess


def is_installed(app):
    """Determine whether program `app` can be found."""
    result = subprocess.run(
        [
            "mdfind",
            "-count",
            f"kMDItemFSName == {app}"
        ],
        capture_output=True,
        text=True,
    )
    if result.returncode > 0 or result.stderr:
        return False
    elif int(result.stdout) >= 1:
        return True



from meety.logging import log


def list_of_ranges(text):
    """Returns a list of ranges induced by a string.

    >>> list_of_ranges("1, 3-5")
    [['1'], ['3', '5']]
    """
    if text and isinstance(text, str):
        ranges = text.split(",")
        return [text_range(r) for r in ranges]
    return []


def text_range(text):
    """Split range of non-dash entries into list of stripped texts.
    Special care is taken on ranges of dates (which contain dashes).

    >>> text_range("foo")
    ['foo']
    >>> text_range("foo - bar")
    ['foo', 'bar']
    >>> text_range("foo-bar-baz")
    ['foo-bar-baz']
    >>> text_range("2020-01-21 - 2020-04-21")
    ['2020-01-21', '2020-04-21']
    """
    count = text.count("-")
    if count == 0:
        return [text.strip()]
    if count == 1:
        return [t.strip() for t in text.split("-")]
    elif count == 5:
        # date ranges:
        parts = [t.strip() for t in text.split("-")]
        return [
            "-".join(parts[0:3]),
            "-".join(parts[3:6])
        ]
    else:
        return [text]


def to_pair(entries, default=(None, None)):
    """Converts a list into a pair if it has at most two entries.
    Otherwise the default pair is returned.

    >>> to_pair([1])
    (1, 1)
    >>> to_pair([1, 2])
    (1, 2)
    >>> to_pair([1, 2, 3])
    (None, None)
    >>> to_pair([1, 2, 3], (8,9))
    (8, 9)
    """
    if len(entries) == 1:
        # make start point also end point
        return (entries[0], entries[0])
    if len(entries) == 2:
        return (entries[0], entries[1])
    else:
        log.expected("pair", "", entries)
        return default



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import (
    QCheckBox,
    QHBoxLayout,
    QWidget,
)


class OptionsWidget(QWidget):
    rating_mode_toggled = pyqtSignal(bool, name="ratingModeToggled")

    def __init__(self):
        super().__init__()
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()

    def _init_style(self):
        pass

    def _add_widgets(self):
        layout = self._create_layout()
        layout.addStretch(1)
        layout.addWidget(self._create_rating_mode())
        self.setLayout(layout)

    def _create_layout(self):
        layout = QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        return layout

    def _create_rating_mode(self):
        self._rating_mode = QCheckBox("only &matching")
        self._rating_mode.setToolTip(" ".join([
            "Show only meetings with matching time or",
            "query constraints. Hide others.",
        ]))
        return self._rating_mode

    def _connect_widget_signals(self):
        self._rating_mode.toggled.connect(self.rating_mode_toggled.emit)



from meety.io.actions import registry
from meety.io.actions.action import Action


class StripAction(Action):
    """Strips string in `texts` from attribute `key` in dictionary `d`.

    For example, `strip_action({1: 'foo-bar', 2: 'baz'}, 1, ['-'])`
    changes the dictionary to {1: 'foobar', 2: 'baz'}.
    """

    name = "strip"

    def __init__(self, key, texts):
        super().__init__()
        self._input = key
        if isinstance(texts, str):
            self._texts = [texts]
        else:
            self._texts = texts

    def is_applicable(self, data, current_level):
        return (
            super().is_applicable(data, current_level)
            and self._input in data.keys()
        )

    def _apply(self, data):
        value = data[self._input]
        for text in self._texts:
            value = value.replace(text, "")
        data[self._input] = value
        return data

    @classmethod
    def create(cls, data):
        input_ = data["input"]
        texts = data["texts"]
        return StripAction(input_, texts)

    @classmethod
    def register(cls):
        registry.register_action_class("strip", StripAction)



"""Windows implementation for system-dependent functionality."""

import subprocess


def is_installed(app):
    """Determine whether program `app` can be found."""
    result = subprocess.run(
        ['where', app],
        capture_output=True,
        text=True,
    )
    return result.returncode == 0



from PyQt5 import QtCore
from PyQt5.QtWidgets import QLabel

from meety.__main__ import (
    SUMMARY,
    URL,
)

ABOUT = f"""<b>Meety - quickly start online meetings from YAML</b><br/>
{SUMMARY}<br/>
<br/>
<a href="{URL}">{URL}</a><br/>
<br/>
Copyright The Contributors to the <i>Meety</i> project.
"""


class TabAbout(QLabel):
    def __init__(self):
        super().__init__()
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()

    def _init_style(self):
        pass

    def _add_widgets(self):
        self.setText(ABOUT)
        self.setMargin(10)
        self.setAlignment(QtCore.Qt.AlignTop)

    def _connect_widget_signals(self):
        pass



"""The graphical user interface program."""

from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QApplication

from meety import connect
from meety.__main__ import PROGRAM
from meety.gui.args import add_argparser_arguments  # USED
from meety.gui.icons import get_icon
from meety.gui.main_window import MainWindow
from meety.io import copy_to_clipboard
from meety.logging import log
from meety.meetings.rating import Ratings

name = "gui"
last_query = []
matching_only = False


def run(args, loader):
    global _loader
    _loader = loader

    _save_args(args)
    _save_meetings(loader.meetings)
    _update_meetings()

    app = _create_app()
    _create_window()
    _on_query_changed("")

    window.show()
    app.exec_()


def _save_args(args):
    global _args
    _args = args


def _save_meetings(meetings):
    global _all_meetings
    _all_meetings = meetings


def get_all_meetings():
    return _all_meetings


def _create_app():
    app = QApplication([])
    icon = get_icon(PROGRAM)
    app.setApplicationName(PROGRAM)
    app.setApplicationDisplayName(PROGRAM)
    app.setWindowIcon(icon)
    return app


def _create_window():
    global window
    window = MainWindow(_args.window_width, _args.window_height)
    _connect_signals()


def _update_meetings():
    global rating
    rating = Ratings(_all_meetings)


def _connect_signals():
    window.query_changed.connect(_on_query_changed)
    window.rating_mode_toggled.connect(_on_rating_mode_toggled)
    window.meeting_chosen.connect(_on_meeting_chosen)
    window.handler_chosen.connect(_on_handler_chosen)
    window.reload_requested.connect(_on_reload_requested)

    global timer
    timer = QTimer()
    timer.timeout.connect(_on_time_changed)
    timer.start(1000)


def _on_reload_requested():
    _loader.reload()
    global _all_meetings
    _all_meetings = _loader.meetings
    _update_meetings()


def _on_time_changed():
    _update_rated_meetings()


def _on_query_changed(text):
    global last_query
    last_query = text.split(" ")
    _update_rated_meetings()


def _on_rating_mode_toggled(status):
    global matching_only
    matching_only = status
    global window
    window.set_meetings_name({
        False: "All meetings",
        True: "Matching meetings",
    }[status])
    _update_rated_meetings()


def _on_meeting_chosen(meeting):
    window.clear_notifications()
    if _args.copy_password:
        _copy_password(meeting)
    handler = _get_handler(meeting)
    _on_handler_chosen((handler, meeting))


def _on_handler_chosen(handler_meeting_pair):
    handler, meeting = handler_meeting_pair
    if handler:
        connect_with_handler(handler, meeting)
    else:
        window.notify("Sorry, no connection handler found!")


def _update_rated_meetings():
    rated_meetings = _rate()
    window.update_rated_meetings(rated_meetings)


def _rate():
    rating.rate(last_query)
    if matching_only:
        return rating.get_all_matching()
    else:
        return rating.get_all()


def _copy_password(meeting):
    password = meeting.data.get("password")
    if password:
        window.notify_quickly("...Copied password to clipboard.")
        copy_to_clipboard(password)


def _get_handler(meeting):
    handlers = connect.applicable_handlers(meeting)
    log.debug(f"Handlers found: {handlers}")
    if not handlers:
        return None
    handler = handlers[0]
    log.debug(f"Handler chosen: {handler}")
    return handler


def connect_with_handler(handler, meeting):
    error = connect.start(handler)
    if error:
        window.notify(error)
    else:
        msg = " ".join([
            f"Started meeting <i>{meeting.name}</i>",
            f"with handler <i>{handler.name}</i>.",
        ])
        window.notify(msg, handler.cmd)



def expect_dict(datum, otherwise={}):
    if isinstance(datum, dict):
        return datum
    return otherwise


def expect_list(datum, otherwise=[]):
    if isinstance(datum, list):
        return datum
    return otherwise


def expect_list_of_str(datum, otherwise=[]):
    if isinstance(datum, str):
        return [datum]
    elif isinstance(datum, list):
        return [str(d) for d in datum]
    return otherwise


def str_is_true(text):
    clean_text = text.strip().casefold()
    return clean_text in ["yes", "true", "on"]


def str_is_false(text):
    return not str_is_true(text)


def replace_words_by(text, words, replacement):
    """
    Replace all given words by the replacement text.

    >>> replace_words_by('Monday and Friday', [' and '], ', ')
    'Monday, Friday'
    >>> replace_words_by('Monday und Friday', [' and '], ', ')
    'Monday und Friday'
    """
    for w in words:
        text = text.replace(w, replacement)
    return text



"""Configuration for the graphical user interface."""

from meety.logging import log

STYLES = {}


def set_options(config):
    if not isinstance(config, dict):
        log.expected("dict", "cli", config)
    else:
        for k, v in config.items():
            _add_style(k, v)


def _add_style(style_name, style_def):
    if not isinstance(style_def, dict):
        log.expected("dict", "styles", style_def)
        return

    STYLES[style_name] = {
        key: style_def.get(key)
        for key in ["color", "background", "selected", "attrs"]
    }


def get_style(style_name):
    return STYLES.get(style_name) or {}


def get_color(style_name="others"):
    return get_style(style_name).get("color") or "black"


def get_background(style_name="others"):
    return get_style(style_name).get("background") or "white"


def get_selected(style_name="others"):
    return get_style(style_name).get("selected") or "white"


def get_attrs(style_name="others"):
    return get_style(style_name).get("attrs") or []


def is_bold(style_name="others"):
    return "bold" in get_attrs(style_name)


def get_border_color():
    return "#aaa"



from meety.logging import log

_action_classes = {}
_actions = []


def reset_options():
    global _actions
    _actions = []


def set_options(actions):
    if not isinstance(actions, list):
        log.expected("list", "actions", actions)
    else:
        for data in actions:
            try_to_add_action(data)


def register_action_class(name, action_class):
    log.debug(f"Register action class '{action_class.name}'.")
    _action_classes[name] = action_class


def get_action_class(data):
    action_name = data.get("action")
    if not action_name:
        log.expected("action name", "attribute 'action'", data)
    return _action_classes.get(action_name)


def get_actions():
    return _actions


def try_to_add_action(data):
    action_class = get_action_class(data)
    if not action_class:
        log.warning(f"Unknown action '{data}'.")
    else:
        _try_to_add_action_for_known_class(
            action_class,
            data
        )


def _try_to_add_action_for_known_class(action_class, data):
    try:
        action = action_class.create(data)
    except Exception as e:
        name = action_class.name
        log.warning(f"Cannot add {name} action '{data}'.")
        log.warning(str(e))
    else:
        action.set_conditions_from_dict(data)
        _actions.append(
            lambda d, level:
            action.apply(d, level)
        )



from PyQt5.QtWidgets import QPlainTextEdit


class TabMeetings(QPlainTextEdit):
    def __init__(self):
        super().__init__()
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()

    def _init_style(self):
        pass

    def _add_widgets(self):
        super().setReadOnly(True)
        self._update_meetings()

    def _update_meetings(self):
        from meety.gui.app import get_all_meetings
        meetings = get_all_meetings()
        text = "\n".join([m.debug_info() for m in meetings])
        if not text:
            text = "There are no meetings!"
        super().setPlainText(text)

    def _connect_widget_signals(self):
        pass



from PyQt5.QtCore import (
    QMargins,
    Qt,
)
from PyQt5.QtGui import (
    QBrush,
    QColor,
    QFont,
    QPen,
)
from PyQt5.QtWidgets import (
    QItemDelegate,
    QStyle,
)

from meety.gui import styles


class ItemDelegate(QItemDelegate):
    """Draw meeting items according to their match status and the GUI
    configuration.
    """

    def __init__(self, parent=None, *args):
        QItemDelegate.__init__(self, parent, *args)

    @classmethod
    def paint(cls, painter, option, index):
        painter.save()
        cls._draw_background(painter, option, index)
        cls._draw_border(painter, option)
        cls._draw_text(painter, option, index)
        painter.restore()

    def sizeHint(self, option, index):
        orig_size = super().sizeHint(option, index)
        return orig_size.grownBy(QMargins(5, 10, 5, 15))

    @classmethod
    def _draw_background(cls, painter, option, index):
        painter.setPen(QPen(Qt.NoPen))
        cls._set_bg_color(painter, option, index)
        painter.drawRect(option.rect)

    @classmethod
    def _set_bg_color(cls, painter, option, index):
        rmeeting = index.data(Qt.UserRole)
        if option.state & QStyle.State_Selected:
            color = styles.get_selected(rmeeting.match_style)
        else:
            color = styles.get_background(rmeeting.match_style)
        painter.setBrush(QBrush(QColor(color)))

    @classmethod
    def _draw_border(cls, painter, option):
        painter.setPen(QPen(QColor(styles.get_border_color())))
        painter.drawLine(
            option.rect.bottomLeft(),
            option.rect.bottomRight()
        )

    @classmethod
    def _draw_text(cls, painter, option, index):
        rmeeting = index.data(Qt.UserRole)
        title = rmeeting.meeting.name
        rect = option.rect.marginsRemoved(QMargins(5, 5, 5, 0))
        cls._set_font_properties(painter, rmeeting)
        painter.drawText(rect, Qt.AlignLeft, title)

    @classmethod
    def _set_font_properties(cls, painter, rmeeting):
        cls._set_font_color(painter, rmeeting)
        cls._set_font_weight(painter, rmeeting)

    @classmethod
    def _set_font_color(cls, painter, rmeeting):
        color = styles.get_color(rmeeting.match_style)
        painter.setPen(QPen(QColor(color)))

    @classmethod
    def _set_font_weight(cls, painter, rmeeting):
        font = QFont()
        font.setBold(styles.is_bold(rmeeting.match_style))
        painter.setFont(font)



"""System-dependent functionality."""

import platform

from meety.logging import log

SYSTEM = platform.system()
_running = None
log.debug(f"Your system is '{SYSTEM}'")
if SYSTEM == "Linux":
    from meety.system import linux
    _running = linux
elif SYSTEM == "Darwin":
    from meety.system import darwin
    _running = darwin
elif SYSTEM == "Windows":
    from meety.system import windows
    _running = windows
else:
    log.warning(f"System '{SYSTEM}' is unknown.")


def satisfies(system, installed):
    """Tests whether the app runs on the given system and whether the
    given program is installed."""
    return all([
        not system or SYSTEM.lower() == system.lower(),
        not installed or _running.is_installed(installed),
    ])



import datetime


class RatedMeeting:
    """A meeting with its time preference and query ratings."""

    PREFER_FACTOR = 1
    QUERY_FACTOR = 10

    def __init__(self, meeting):
        """Save meeting with preference and query rating."""
        self._meeting = meeting
        self._prating = False
        self._qrating = False

    @property
    def meeting(self):
        return self._meeting

    def set_preference_rating(self, prating):
        self._prating = prating

    def set_query_rating(self, qrating):
        self._qrating = qrating

    @property
    def rating(self):
        """Compute total rating value."""
        return sum([
            RatedMeeting.PREFER_FACTOR * self._prating,
            RatedMeeting.QUERY_FACTOR * self._qrating,
        ])

    def matches_query(self):
        return self._qrating

    def matches_preference(self):
        return self._prating

    def matches(self):
        return self.rating > 0

    @property
    def match_style(self):
        if self.matches_query():
            return "matching_query"
        if self.matches_preference():
            return "matching_preference"
        return "others"

    def __str__(self):
        return str(self._meeting)

    def debug_info(self):
        name = str(self._meeting)
        if self.matches_query():
            status = "Q"
        elif self.matches_preference():
            status = "T"
        else:
            status = "x"
        return f"{self.rating:3d} {status} {name}"


class Ratings:
    """All meetings and their ratings."""

    @classmethod
    def rate_meeting(cls, meeting, iquery, when):
        rmeeting = RatedMeeting(meeting)
        rmeeting.set_preference_rating(meeting.match_time(when))
        rmeeting.set_query_rating(meeting.match_query(iquery))
        return rmeeting

    def __init__(self, meetings):
        """Start without ratings."""
        self._meetings = meetings
        self._rated_meetings = []

    def rate(self, query, when=None):
        """Rate all meetings based on the `query` and the provided
        datetime `when` (or current time, by default).
        """
        if when is None:
            when = datetime.datetime.now()
        matches = self._rate_meetings(query, when)
        self._rated_meetings = sorted(
            matches,
            key=lambda m: m.rating,
            reverse=True,
        )

    def _rate_meetings(self, query, when):
        return [
            Ratings.rate_meeting(m, query, when)
            for m in self._meetings
        ]

    def get_all(self):
        """Return all meetings."""
        return self._rated_meetings

    def get_all_matching(self):
        """Return only meetings with a positive rating."""
        return [rm for rm in self._rated_meetings if rm.rating > 0]

    def get_only_best(self):
        """Return only meetings with a maximal rating."""
        if self._rated_meetings:
            best_rating = self._rated_meetings[0].rating
            return [
                rm for rm in self._rated_meetings
                if rm.rating >= best_rating
            ]

    def get_only_first(self):
        """Get only the first meeting (with a maximal rating)."""
        if self._rated_meetings:
            return [self._rated_meetings[0]]



import os
import re
from setuptools import setup
from setuptools.config import read_configuration


with open("src/meety/__main__.py", encoding="utf8") as f:
    version = re.search(r'VERSION = "(.*?)"', f.read()).group(1)

config_file = os.path.join(os.path.dirname(__file__), "setup.cfg")
config = read_configuration(config_file)

setup(
    **config["metadata"],
    **config["options"],
    package_data={
        "meety": [
            os.path.join("static", "config", "*.yaml"),
            os.path.join("static", "icons", "*.png"),
            os.path.join("gui", "static", "*.css"),
        ]
    },
    version=version,
)



from PyQt5.QtCore import (
    QEvent,
    Qt,
    pyqtSignal,
)
from PyQt5.QtWidgets import (
    QGridLayout,
    QLabel,
    QListWidget,
    QWidget,
)

from meety.gui.main_window.meeting_list.context import MeetingItemMenu
from meety.gui.main_window.meeting_list.delegate import ItemDelegate
from meety.gui.main_window.meeting_list.item import MeetingItem
from meety.gui.main_window.options import OptionsWidget


class MeetingList(QWidget):
    meeting_chosen = pyqtSignal(object, name="meeting_chosen")
    handler_chosen = pyqtSignal(object, name="handler_chhosen")
    reload_requested = pyqtSignal(name="reload_requested")
    rating_mode_toggled = pyqtSignal(bool, name="rating_mode_toggled")

    def __init__(self):
        super().__init__()
        self.name = "Meetings"
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()
        self.update_rated_meetings([])

    def _init_style(self):
        pass

    def _add_widgets(self):
        layout = self._create_layout()
        layout.addWidget(self._create_info(), 0, 0)
        layout.addWidget(self._create_options(), 0, 1)
        layout.addWidget(self._create_list(), 1, 0, 1, 2)
        self.setLayout(layout)

    def _create_layout(self):
        layout = QGridLayout()
        layout.setRowStretch(0, 0)
        layout.setRowStretch(1, 1)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        return layout

    def _create_info(self):
        self._info = QLabel("Meetings")
        self._info.setStyleSheet("""
            font: bold;
            margin-top: 5px;
        """)
        return self._info

    def _create_options(self):
        self._options = OptionsWidget()
        return self._options

    def _create_list(self):
        self._list = QListWidget()
        item_delegate = ItemDelegate()
        self._list.setItemDelegate(item_delegate)
        return self._list

    def _connect_widget_signals(self):
        self._list.itemActivated.connect(self._choose_meeting)
        self._options.rating_mode_toggled.connect(
            self.rating_mode_toggled.emit
        )
        self._list.installEventFilter(self)

    def set_name(self, name):
        self.name = name
        self._update_info()

    def update_rated_meetings(self, rated_meetings):
        self._rated_meetings = rated_meetings
        self._update_info()
        self._update_entries(rated_meetings)

    def _update_info(self):
        num = len(self._rated_meetings)
        self._info.setText(f"{self.name} ({num})")

    def _update_entries(self, rated_meetings):
        row = self._list.currentRow()
        last = len(rated_meetings) - 1
        self._list.clear()
        for rmeeting in rated_meetings:
            item = MeetingItem(rmeeting)
            self._list.addItem(item)
        self._list.setCurrentRow(min(row, last))

    def _choose_meeting(self, item):
        rmeeting = item.data(Qt.UserRole)
        self.meeting_chosen.emit(rmeeting.meeting)

    def eventFilter(self, source, event):
        if event.type() == QEvent.ContextMenu and source is self._list:
            return self._context_menu_on_list(source, event)
        return super().eventFilter(source, event)

    def _context_menu_on_list(self, source, event):
        item = source.itemAt(event.pos())
        rmeeting = item.data(Qt.UserRole)
        menu = MeetingItemMenu(rmeeting.meeting)
        menu.handler_chosen.connect(self.handler_chosen.emit)
        menu.reload_requested.connect(self.reload_requested.emit)
        menu.exec_(event.globalPos())
        return True



"""Meety - quickly connect to online meetings managed in YAML."""

import argparse
import sys

from meety import args as common_args
from meety.config import Config
from meety.loader import Loader
from meety.logging import log

PROGRAM = "meety"
VERSION = "0.9.0"
SUMMARY = f"This is {PROGRAM}, version {VERSION}."
URL = "https://github.com/GaetanoGeck/meety/"


def start_cli():
    try:
        from meety.cli import app
        start_app(app)
    except KeyboardInterrupt:
        print("Goodbye!")


def start_gui():
    from meety.gui import app
    start_app(app)


def start_app(app):
    config = _prepare_config(PROGRAM)
    args = _prepare_args(app, config)
    _prepare_logger(args)
    config.reload()

    config.apply()
    loader = _load(args)
    app.run(args, loader)


def _prepare_config(program_name):
    Config.PROGRAM = program_name
    config = Config()
    return config


def _prepare_args(app, config):
    argparser = _setup_argparser(app)
    return _combine_args(app, config, argparser)


def _setup_argparser(app):
    argparser = argparse.ArgumentParser(
        prog=sys.argv[0],
        description=__doc__,
    )
    app.add_argparser_arguments(argparser)
    common_args._add_argparser_common_arguments(
        argparser,
        SUMMARY
    )
    return argparser


def _combine_args(app, config, argparser):
    user_args = config.user.args(app.name) or []
    args = user_args + sys.argv[1:]
    return argparser.parse_args(args)


def _prepare_logger(args):
    log.setWarn()
    if args.verbose:
        log.setInfo()
    if args.debug:
        log.setDebug()


def _load(args):
    loader = Loader()
    loader.only_explicit = args.only_explicit
    loader.add_explicit_directories(args.directories)
    loader.add_explicit_files(args.files)
    loader.load()
    return loader


if __name__ == "__main__":
    start_cli()



from contextlib import suppress

from meety.io.actions import registry
from meety.io.actions.action import Action


class InferAction(Action):
    """Adds entry `name` to dictionary if not already existent and if its
    value can be computed.

    For example, `infer_action({"id": "JohnDoe"}, "mail", "{id}@mail.com")`
    adds the key "mail" with value "JohneDoe@mail.com" to the dictionary.
    """

    name = "infer"

    def __init__(self, name, value):
        super().__init__()
        self._name = name
        self._value = value

    def is_applicable(self, data, current_level):
        return (
            super().is_applicable(data, current_level)
            and self._name not in data.keys()
        )

    def _apply(self, data):
        with suppress(KeyError):
            expanded = self._value.format_map(data)
            data[self._name] = expanded
        return data

    @classmethod
    def create(cls, data):
        name = data["name"]
        value = data["value"]
        return InferAction(name, value)

    @classmethod
    def register(cls):
        registry.register_action_class("infer", InferAction)
